from ..exception import InvalidAttributeError

from .mongo import MongoRepository


def get_repository(index_name, repo_type):
    repo = MongoRepository(index_name)
    return RepositoryWrapper(repo)


class RepositoryWrapper:
    def __init__(self, repo):
        self.db = repo

    def clean(self):
        self.db.clean()

    def insert(self, data):
        if isinstance(data, dict):
            return self.db.insert_one(data)

        if isinstance(data, list):
            return self.db.insert_many(data)

        raise InvalidAttributeError(
            data, 'data of type dict or list expected, got {}'.format(type(data)))

    def find(self, query, many=False):
        if many:
            return self.db.find(query)
        else:
            return self.db.find_one(query)



from __future__ import division

import os
import sys
import torch
import torchvision.transforms as transforms
from PIL import Image

from .models import Darknet
from .utils import load_classes, pad_to_square, resize, non_max_suppression, rescale_boxes, download_files

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
DATA_PATH = os.path.join(os.environ.get('HOME'), '.imsearch', 'yolo')


class Detector(object):
    model = None

    def __init__(self):
        download_files()

        self.img_size = 416
        self.config_path = os.path.join(DATA_PATH, 'yolov3.cfg')
        self.weights_path = os.path.join(DATA_PATH, 'yolov3.weights')
        self.class_path = os.path.join(DATA_PATH, 'coco.names')
        self.conf_thres = 0.8
        self.nms_thres = 0.4
        if Detector.model is None:
            model = Darknet(self.config_path,
                            img_size=self.img_size).to(device)
            model.load_darknet_weights(self.weights_path)
            model.eval()
            Detector.model = model
        self.classes = load_classes(self.class_path)

    def _load_image(self, img):
        actual_size = img.shape
        img = self._process_image(img)
        img = img.unsqueeze_(0)
        return img, actual_size
    
    def _process_image(self, img):
        img = Image.fromarray(img)
        img = transforms.ToTensor()(img)
        img, _ = pad_to_square(img, 0)
        img = resize(img, self.img_size)
        return img

    def predict(self, img):
        input_img, actual_shape = self._load_image(img)
        input_img = input_img.to(device)

        with torch.no_grad():
            detections = Detector.model(input_img)
            detections = non_max_suppression(
                detections, self.conf_thres, self.nms_thres)[0]

        response = []
        if detections is not None:
            detections = rescale_boxes(
                detections, self.img_size, actual_shape[:2])
            for x1, y1, x2, y2, _, cls_conf, cls_pred in detections:
                x1 = max(min(int(x1.item()), actual_shape[1]-1), 0)
                x2 = max(min(int(x2.item()), actual_shape[1]-1), 0)
                y1 = max(min(int(y1.item()), actual_shape[0]-1), 0)
                y2 = max(min(int(y2.item()), actual_shape[0]-1), 0)
                response.append({
                    'box': [x1, y1, x2, y2],
                    'score': cls_conf.item(),
                    'label': int(cls_pred),
                    'name': self.classes[int(cls_pred)]
                })
        return response



import os
import subprocess

dir_path = os.path.dirname(os.path.realpath(__file__))


def run(redis_url=None):
    env = os.environ.copy()
    if redis_url is not None:
        env['REDIS_URI'] = redis_url
    return subprocess.Popen(['python', os.path.join(dir_path, 'extractor.py')], env=env)



import os
import sys
import shutil
import base64
import uuid
import json
import time
from PIL import Image
import numpy as np
import redis

from . import utils


class FeatureExtractor:
    def __init__(self, index_name):
        self.index_name = index_name
        self.redis_url = os.environ.get('REDIS_URI')
        self._redis_db = redis.StrictRedis.from_url(self.redis_url)
        os.makedirs(self._get_image_path(), exist_ok=True)

    def _get_image_path(self):
        home_dir = os.environ.get('HOME')
        return os.path.join(home_dir, '.imsearch', 'images', self.index_name)

    def _save_image(self, image, _id):
        dst = os.path.join(self._get_image_path(), '{}.jpg'.format(_id))
        utils.save_image(image, dst)
        return dst

    def _decode_redis_data(self, data):
        data = json.loads(data.decode('utf-8'))

        def _decode(d):
            d['features'] = utils.base64_decode(
                d['features'], dtype=np.float32)
            return d

        data['primary'] = list(map(_decode, data['primary']))
        data['secondary'] = utils.base64_decode(
            data['secondary'], dtype=np.float32)
        return data

    def clean(self):
        shutil.rmtree(self._get_image_path())
        os.makedirs(self._get_image_path(), exist_ok=True)

    def extract(self, image_path, save=True):
        image = utils.check_load_image(image_path)
        if image is None:
            return None
        _id = str(uuid.uuid4())
        data = {
            'id': _id,
            'image': utils.base64_encode(image),
            'shape': image.shape
        }
        self._redis_db.rpush('image_queue', json.dumps(data))

        result = None
        while result is None:
            time.sleep(0.01)
            result = self._redis_db.get(_id)

        result = self._decode_redis_data(result)
        result['id'] = _id
        if save:
            result['image'] = self._save_image(image, _id)

        if 'http' in image_path:
            result['url'] = image_path

        self._redis_db.delete(_id)
        return result



"""
index.py
====================================
The module contains the Index class 
"""

import os
import numpy as np
import copy
import requests

from .nmslib import NMSLIBIndex
from .extractor import FeatureExtractor
from .repository import get_repository

from .backend import run


class Index:
    """
    The class to create the searchable index object.
    """
    fe = None
    object_counter = 0

    def __init__(self, name):
        """
        Create the index with its name.
        Parameters
        ---------
        name
            Unique indentifier for your searchable index object. 
        """
        Index.object_counter += 1
        if os.environ.get('DETECTOR_MODE') == 'local' and (Index.fe is None or Index.fe.poll() is not None):
            Index.fe = run()
        self.match_ratio = 0.3
        self.name = name
        self._nmslib_index = NMSLIBIndex(self.name)
        self._feature_extractor = FeatureExtractor(self.name)
        self._repository_db = get_repository(self.name, 'mongo')

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._nmslib_index.createIndex()

    def __del__(self):
        Index.object_counter -= 1
        if Index.object_counter == 0 and Index.fe is not None:
            Index.fe.terminate()

    def _get_object_wise_similar(self, features):
        matches = {}
        for data in features['primary']:
            knn = self._nmslib_index.knnQuery(
                data['features'], 'primary', k=10)
            for x in knn:
                img_data = self._repository_db.find({
                    'primary': {
                        'index': x[0],
                        'label': data['label'],
                        'name': data['name']
                    }
                })
                if img_data is not None:
                    _id = img_data['_id']

                    if _id in matches:
                        matches[_id]['s_dist'] = x[1] + matches[_id]['s_dist']
                    else:
                        matches[_id] = {
                            'data': copy.deepcopy(img_data),
                            's_dist': x[1]
                        }

        knn = self._nmslib_index.knnQuery(
            features['secondary'], 'secondary', k=10)
        for x in knn:
            img_data = self._repository_db.find({'secondary_index': x[0]})
            if img_data is not None:
                _id = img_data['_id']
                if _id in matches:
                    matches[_id]['p_dist'] = x[1]
                else:
                    matches[_id] = {
                        'data': copy.deepcopy(img_data),
                        'p_dist': x[1]
                    }

        matches = list(matches.values())
        total_objects = float(len(features['primary']))

        def update_scores(data):
            score = (1 - self.match_ratio)*data.get('p_dist', 0) / \
                (total_objects + 10e-5) + self.match_ratio*data.get('s_dist', 0)
            return (data['data'], score)

        matches = list(map(update_scores, matches))
        matches.sort(key=lambda x: x[1])
        return matches

    def cleanIndex(self):
        """
        Cleans the index. It will delete the images already added to the index. It will also remove the database entry for the same index.
        """
        self._feature_extractor.clean()
        self._nmslib_index.clean()
        self._repository_db.clean()

    def addImage(self, image_path):
        """
        Add a single image to the index. 
        Parameters
        ---------
        image_path
            The local path or url to the image to add to the index.
        """
        features = self._feature_extractor.extract(image_path)
        if features is None:
            return False
        reposiory_data = self._nmslib_index.addDataPoint(features)
        self._repository_db.insert(reposiory_data)
        return True

    def addImageBatch(self, image_list):
        """
        Add a multiple image to the index. 
        Parameters
        ---------
        image_list
            The list of the image paths or urls to add to the index.
        """
        response = []
        for image_path in image_list:
            response.append(self.addImage(image_path))
        return response

    def createIndex(self):
        """
        Creates the index. Set create time paramenters and query-time parameters for nmslib index. 
        """
        self._nmslib_index.createIndex()

    def knnQuery(self, image_path, k=10, policy='global'):
        """
        Query the index to search for k-nearest images in the database.
        Parameters
        ---------
        image_path
            The path to the query image.
        k=10
            Number of results.
        policy='global'
            choose policy from 'object' or 'global'. Search results will change accordingly.
            object: Object level matching. The engine will look for similarity at object level for every object detected in the image.
            global: Overall similarity using single feature space on the whole image.
        """

        features = self._feature_extractor.extract(image_path, save=False)
        matches = []
        if policy == 'object':
            matches = self._get_object_wise_similar(features)

        if not matches:
            knn = self._nmslib_index.knnQuery(
                features['secondary'], 'secondary', k=10)
            matches = [(self._repository_db.find(
                {'secondary_index': x[0]}, many=False), 1.0/x[1]) for x in knn]

        return matches[:k]



from .config import config_init
from .index import Index
from .backend import run


def init(name='index',**config):
    config_init(config)
    return Index(name=name)

def run_detector(redis_url):
    run(redis_url)


import os
import copy

default_config = {
    'MONGO_URI': 'mongodb://localhost:27017/',
    'REDIS_URI': 'redis://localhost:6379/0',
    'DETECTOR_MODE': 'local'
}


def config_init(config):
    final_config = copy.deepcopy(default_config)
    for k, v in config.items():
        final_config[k] = v
    os.environ.update(final_config)



import os
import sys
import time
import json
import base64
import numpy as np
import redis

from object_detector import get_detector
from feature_extractor import extract_features
from imsearch import utils

REDIS_DB = redis.StrictRedis.from_url(
    os.environ.get('REDIS_URI', 'redis://localhost:6379/0'))
REDIS_QUEUE = 'image_queue'
BATCH_SIZE = 1


def main():
    detector = get_detector('yolo')
    while True:
        time.sleep(0.01)
        _queue = REDIS_DB.lrange(REDIS_QUEUE, 0, BATCH_SIZE - 1)
        for _q in _queue:
            all_features = {
                'primary': [],
                'object_bitmap': [0 for _ in range(len(detector.classes))]
            }
            _q = json.loads(_q.decode("utf-8"))
            img = utils.base64_decode(_q['image'], _q['shape'])

            all_features['secondary'] = extract_features(img.copy())

            response = detector.predict(img)
            for obj in response:
                box = obj['box']
                x1, y1, x2, y2 = box[0], box[1], box[2], box[3]
                if(x2-x1 >= 75 and y2-y1 >= 75):
                    features = extract_features(img[y1:y2, x1:x2])
                    all_features['primary'].append({
                        'features': features,
                        'label': obj['label'],
                        'name': obj['name'],
                        'box': obj['box']
                    })
                    all_features['object_bitmap'][obj['label']] = 1

            REDIS_DB.set(_q['id'], json.dumps(all_features))
        REDIS_DB.ltrim(REDIS_QUEUE, len(_queue), -1)


if __name__ == "__main__":
    print("Running extractor")
    main()



import os
from pymongo import MongoClient


class MongoRepository:
    def __init__(self, index_name):
        url = os.environ.get('MONGO_URI')
        self.db = MongoClient(url).imsearch[index_name]

    def clean(self):
        self.db.drop()

    def insert_one(self, data):
        return self.db.insert_one(data)

    def insert_many(self, data):
        return self.db.insert_many(data)

    def find_one(self, query):
        response = self.db.find_one(query)
        return response

    def find(self, query):
        return list(self.db.find(query))



import os
import shutil
import pandas as pd
import numpy as np
import nmslib

instance_map = {}


def NMSLIBIndex(index_name):
    if index_name not in instance_map:
        instance_map[index_name] = _nmslibIndex(index_name)
    return instance_map[index_name]


class _nmslibIndex:
    def __init__(self, name):
        self.index_name = name
        self._load_index()

    def _add_data(self, index, df):
        data = df.values.astype(np.float32)
        c = index.addDataPointBatch(data, range(data.shape[0]))
        return index, c

    def _get_index_path(self):
        home_dir = os.environ.get('HOME')
        return os.path.join(home_dir, '.imsearch', 'indices', self.index_name, 'index.h5')

    def _load_index(self):
        index_file = self._get_index_path()

        self.primary = nmslib.init(
            method='hnsw', space='l2', data_type=nmslib.DataType.DENSE_VECTOR)
        self.secondary = nmslib.init(
            method='hnsw', space='l2', data_type=nmslib.DataType.DENSE_VECTOR)
        self.bitmap = nmslib.init(
            method='hnsw', space='l2', data_type=nmslib.DataType.DENSE_VECTOR)

        if os.path.exists(index_file):
            self.primary_df = pd.read_hdf(index_file, 'primary')
            self.primary, self.primary_c = self._add_data(
                self.primary, self.primary_df)

            self.secondary_df = pd.read_hdf(index_file, 'secondary')
            self.secondary, self.secondary_c = self._add_data(
                self.secondary, self.secondary_df)

            self.bitmap_df = pd.read_hdf(index_file, 'bitmap')
            self.bitmap, self.bitmap_c = self._add_data(
                self.bitmap, self.bitmap_df)
        else:
            self.primary_df = None
            self.secondary_df = None
            self.bitmap_df = None

            self.primary_c, self.secondary_c, self.bitmap_c = 0, 0, 0

    def _add_data_point(self, index, count, data):
        _t = index.addDataPoint(count, data)
        return index, _t + 1

    def clean(self):
        index_file = self._get_index_path()
        if os.path.exists(os.path.dirname(index_file)):
            shutil.rmtree(os.path.dirname(index_file))
            self._load_index()

    def getDataByIds(self, id_list, _type='primary'):
        return [self.getDataById(x, _type) for x in id_list]

    def getDataById(self, _id, _type='primary'):
        return getattr(self, _type + '_df').iloc[_id].values

    def addDataPoint(self, data):
        primary = []
        for d in data['primary']:
            primary.append({
                'index': self.primary_c,
                'label': d['label'],
                'name': d['name'],
                'box': d['box']
            })
            v = d['features']
            if self.primary_df is None:
                self.primary_df = pd.DataFrame(columns=range(v.shape[0]))
            self.primary_df.loc[self.primary_c] = v
            self.primary, self.primary_c = self._add_data_point(
                self.primary, self.primary_c, v)

        db_data = {
            '_id': data['id'],
            'image': data['image'],
            'url': data.get('url', ''),
            'primary': primary,
            'secondary_index': self.secondary_c,
            'bitmap_index': self.bitmap_c
        }

        v = data['secondary']
        if self.secondary_df is None:
            self.secondary_df = pd.DataFrame(columns=range(v.shape[0]))
        self.secondary_df.loc[self.secondary_c] = v
        self.secondary, self.secondary_c = self._add_data_point(
            self.secondary, self.secondary_c, v)

        v = np.array(data['object_bitmap'], dtype=np.float32)
        if self.bitmap_df is None:
            self.bitmap_df = pd.DataFrame(columns=range(v.shape[0]))
        self.bitmap_df.loc[self.bitmap_c] = v
        self.bitmap, self.bitmap_c = self._add_data_point(
            self.bitmap, self.bitmap_c, v)

        return db_data

    def addDataPointBatch(self, data_list):
        return [self.addDataPoint(data) for data in data_list]

    def createIndex(self):
        M = 15
        efC = 100
        num_threads = 4
        index_time_params = {
            'M': M, 'indexThreadQty': num_threads, 'efConstruction': efC}
        self.primary.createIndex(index_time_params)
        self.secondary.createIndex(index_time_params)
        self.bitmap.createIndex(index_time_params)

        os.makedirs(os.path.dirname(self._get_index_path()), exist_ok=True)
        
        if self.primary_df is not None:
            self.primary_df.to_hdf(self._get_index_path(), 'primary')
        
        if self.secondary_df is not None:
            self.secondary_df.to_hdf(self._get_index_path(), 'secondary')
        
        if self.bitmap_df is not None:
            self.bitmap_df.to_hdf(self._get_index_path(), 'bitmap')

        efs = 100
        query_time_params = {'efSearch': efs}
        self.primary.setQueryTimeParams(query_time_params)
        self.secondary.setQueryTimeParams(query_time_params)
        self.bitmap.setQueryTimeParams(query_time_params)

    def knnQuery(self, data, _type='bitmap', k=10):
        response = getattr(self, _type).knnQuery(data, k=k)
        return [(int(i), float(d)) for i, d in zip(response[0], response[1])]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='imsearch',
    version='0.1.1',
    description='A generic framework to build your own reverse image search engine',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/rikenmehta03/imsearch',
    author='Riken Mehta',
    author_email='riken.mehta03@gmail.com',
    packages=setuptools.find_packages(),
    install_requires=['torch', 'torchvision', 'pandas', 'tables', 'redis', 'pymongo', 'nmslib', 'wget', 'opencv-python', 'requests'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX :: Linux",
    ]
)



from PIL import Image
import base64
import torch
import torch.nn as nn
from torchvision import models
import torchvision.transforms as transforms


def get_extractor(arch='resnet50'):
    model_ft = models.__dict__[arch](pretrained=True)
    extractor = nn.Sequential(*list(model_ft.children())[:-1])
    return extractor


device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
extractor = get_extractor('resnet50').to(device)
extractor.eval()
image_transforms = transforms.Compose([
    transforms.Resize(300),
    transforms.CenterCrop(244),
    transforms.ToTensor(),
])


def extract_features(img):
    img = image_transforms(Image.fromarray(img))
    img = img.unsqueeze_(0).to(device)

    with torch.no_grad():
        out = extractor(img).squeeze().detach().cpu()
    return base64.b64encode(out.numpy()).decode("utf-8")



import sys
import base64
import requests
import cv2
from PIL import Image
from io import BytesIO

import numpy as np


def check_load_image(image_path):
    if 'http' in image_path:
        r = requests.get(image_path)
        if r.status_code != 200:
            return None
        img = np.asarray(Image.open(BytesIO(r.content)))
    else:
        img = np.asarray(Image.open(image_path))

    if len(img.shape) != 3:
        return None

    if img.shape[2] > 3:
        img = img[:, :, :3]

    return img


def save_image(image, dest):
    image = Image.fromarray(image)
    image.save(dest)


def base64_encode(a):
    a = a.copy(order='C')
    return base64.b64encode(a).decode("utf-8")


def base64_decode(a, shape=None, dtype=np.uint8):
    if sys.version_info.major == 3:
        a = bytes(a, encoding="utf-8")

    a = np.frombuffer(base64.decodestring(a), dtype=dtype)
    if shape is not None:
        a = a.reshape(shape)
    return a
